# -*- coding:utf-8 -*- 
#
# Copyright(C) 2014 Laurent Faipot (laurent.faipot@free.fr). All rights reserved.
#

from processor import *

class Architecture:

    memorySize      = 256

class Chip(Processor):
    
    wordSize        = 1 # byte
    regSize         = 1 # byte
    instructionSize = 1 # byte
    operandSize     = 1 # byte
    
    def lda(self, board, operand):
        self.regA.set(board.memory.get(operand, 1))
        
    def ldb(self, board, operand):
        self.regB.set(board.memory.get(operand, 1))
        
    def sta(self, board, operand):
        board.memory.set(operand, board.chip.wordSize, self.regA.get())
    
    def stb(self, board, operand):
        board.memory.set(operand, board.chip.wordSize, self.regB.get())
        
    def mova(self, board, operand):
        self.regA.set(operand)
        
    def movb(self, board, operand):
        self.regB.set(operand)
    
    def add(self, board, operand):
        value = self.regA.get() + self.regB.get()
        if operand == 0:
            reg = self.regA
        else:
            reg = self.regB
        reg.set(value)
    
    def dec(self, board, operand):
        if operand == 0:
            reg = self.regA
        else:
            reg = self.regB
        reg.set(reg.get() - 1)
            
    def jmp(self, board, operand):
        self.PC.set(operand)
    
    def jmpz(self, board, operand):
        if (self.regA.get() == 0):
            self.PC.set(operand)
            
    def end(self, board, operand):
        self.endProgram = True

   
    operandSet = [
        [Processor.OpAddrData, "(\d*)", Processor.OpFormatDecimal, -1],
        [Processor.OpAddrCode, "(\d*)", Processor.OpFormatDecimal, -1],
        [Processor.OpNumeric, "#(-?\d*)", Processor.OpFormatDecimal, -1],
        [Processor.OpAccIndic,  "A", Processor.OpFormatRegister, 0],
        [Processor.OpAcc,      "B", Processor.OpFormatRegister, 1]
        ]
        
    instructionSet= {
        'LDA':  [0x01, lda, [Processor.OpAddrData]],
        'LDB':  [0x02, ldb, [Processor.OpAddrData]],
        'STA':  [0x03, sta, [Processor.OpAddrData]],
        'STB':  [0x04, stb, [Processor.OpAddrData]],
        'ADD':  [0x05, add, [Processor.OpAccIndic, Processor.OpAcc]],
        'DEC':  [0x06, dec, [Processor.OpAccIndic, Processor.OpAcc]],
        'JMP':  [0x07, jmp, [Processor.OpAddrCode]],
        'JMPZ': [0x08, jmpz, [Processor.OpAddrCode]],
        'MOVA': [0x09, mova, [Processor.OpNumeric]],
        'MOVB': [0x0A, movb, [Processor.OpNumeric]],
        '.END':  [0xFF, end, [Processor.OpVoid]]
        }
    
        
    def __init__(self):
        #super(Chip, self).__init__("test")
        Processor.__init__(self, "edu1", Chip.instructionSize, Chip.operandSize)
        self.PC    = self.addRegister('PC', Register.PC, Register.UnsignedInteger, Chip.regSize)
        self.regA  = self.addRegister('A', Register.AccIndic, Register.SignedInteger, Chip.regSize)
        self.regB  = self.addRegister('B', Register.Acc, Register.SignedInteger, Chip.regSize)
        
        self.zero       = self.addIndicator('Z', Indicator.Zero)
        
        self.defineOperand(self.operandSet)
        self.defineInstruction(self.instructionSet)
        
    
            
class Program(Assembly):

    def __init__(self):
        Assembly.__init__(self)

    def load(self, fileName, board):
        memory = board.memory
        "load code into memory"
        f = open(fileName, "r");
        dataSection = False
        lineNo = 0
        
        error = False
        errorList = []
        
        for line in f:
            
            lineNo += 1
            
            # .DATA
            #   signale le debut des donnes suivi de ligne
            # <adresse> valeur
            #
            # .CODE <adresse>
            #    signale le debut du code avec adresse de chargement
            # chaque instruction est compose de 2 champs
            # le mnemonique d'instruction et l'operande (sauf END)
            #
            
            # remove comments
            line = line.split(';', 1)[0]
            # split using spaces/tabs as separators
            instruction = line.split()
            
            # check if the line is empty, 
            if len(instruction) == 0:
                # probably an empty line
                continue


            mnemonic = instruction[0].upper()
            operand = None

            # is there an operand
            if (len(instruction) > 1):
                operand   = instruction[1]
            
            if (mnemonic == '.DATA'):
                dataSection = True
                continue
                
            if (mnemonic == '.CODE'):
                dataSection = False
                # we store starting address of code
                if (operand):
                    self.setCodeBase(int(operand))
                else:
                    error = True
                    errorList.append("Line " + str(lineNo) + ": .CODE requires starting address")
                continue

            if (dataSection):
                # format: <address> <value>
                memory.set(int(mnemonic), 1, int(operand))
                continue
            
            address = self.getLoadAddress()
            errorMsg = self.loadAssembly(board, mnemonic, operand)
            if (len(errorMsg) == 0):
                if operand == None:
                    operand = 0
                self.storeInstruction(address, mnemonic, operand)
                
            else:
                errorList.append("Line " + str(lineNo) + ": " + errorMsg)
            
        # add a .END if there is none at the end of the source file
        if (mnemonic != ".END"):
            errorMsg = self.loadAssembly(board, ".END", 0)
            self.storeInstruction(self.loadAddress, ".END", 0)
            
        if (error):
            raise ProgramError(Error.error, errorList)
                    
