# -*- coding:utf-8 -*- 
#
# Copyright(C) 2014 Laurent Faipot (laurent.faipot@free.fr). All rights reserved.
#

import re
from error import *

class ProgramError(Error):
    def __init__(self, severity, msgList):
        Error.__init__(self, severity)
        self.msgList = msgList
        
    def __str__(self):
        msg = ""
        for l in self.msgList:
            if (len(msg) > 0):
                msg += "\n"
            msg += l
        return msg
    
class ExecError(Error):
    def __init__(self, severity, msg):
        Error.__init__(self, severity)
        self.msg = msg
        
    def __str__(self):
        return self.msg

class Indicator:

    Carry       = 1
    Zero        = 2
    Overflow    = 3
    Negative    = 4
    
    __rank      = 0

    @classmethod
    def reset(self):
        Indicator.__rank = 0
        
    @classmethod
    def whenUpdated(cls, fct):
        cls.displayChange = fct
        
    def __init__(self, processor, name, indicatorType):
        self.processor = processor
        self.name = name
        self.indicatorType = indicatorType
        self.value = 0
        self.rank = Indicator.__rank
        Indicator.__rank += 1
    
    def getName(self):
        return self.name
    
    def getRank(self):
        return self.rank
    
    def get(self):
        return self.value
    
    def set(self, value):
        if (value != self.value):
            self.value = value
            Indicator.displayChange(self)
    
class Register:
    
    PC              = 1
    SP              = 2
    Index           = 3
    AccIndic        = 4
    Acc             = 5
    
    SignedInteger    = 1
    UnsignedInteger  = 2
    Float            = 3

    
    __rank = 0
    
    def __init__(self, processor, name, regType, regFormat, size):
        self.processor = processor
        self.name = name
        self.regType = regType
        self.regFormat = regFormat
        self.value = 0
        self.size = size
        if (regFormat == Register.SignedInteger):
            self.min  = - (2 ** (size * 8 - 1))
            self.max  = 2 ** (size * 8 - 1) -1
        elif (regFormat == Register.UnsignedInteger):
            self.min  = 0
            self.max  = 2 ** (size * 8) - 1
        self.rank = Register.__rank
        Register.__rank += 1
    
    @classmethod
    def reset(self):
        Register.__rank = 0
    
    @classmethod    
    def whenUpdated(cls, fct):
        cls.displayChange = fct
       
    def set(self, value):
        # convert in negative value
        overflow = 0
        print 'REG FORMAT' + str(self.regFormat)
        
        if (self.regFormat == Register.SignedInteger):
            mask = 1 << (self.size * 8 - 1)
            if (value & mask):
                # propagate sign bit
                mask = -1 << (self.size * 8)
                value = value | mask
 
        if (self.regFormat in (Register.UnsignedInteger, Register.SignedInteger)):
            print "value: " + str(value) + ' max: ' + str(self.max) + ' min ' + str(self.min)
            if (value > self.max):
                print 'MAX: value: ' + str(value)
                value = (~value) - 1
                value2 = - value
                print '   -> value: ' + str(value) + ' value2 ' + str(value2)
 
                overflow = 1
            elif (value < self.min):
                # 2's complement
                print 'MIN: value: ' + str(value)
                value = (~value) - 1
                print '   -> value: ' + str(value)
                overflow = 1
 
        self.value = value
        Register.displayChange(self)
        if (self.regType == Register.AccIndic):
            self.processor.setZero(value)
            self.processor.setNeg(value)
            self.processor.setOverflow(overflow)

    def get(self):
        return self.value
    
    def getName(self):
        return self.name
    
    def getRank(self):
        return self.rank
    
class Processor:
        
    OpVoid            = 0
    OpAddrCode        = 1
    OpAddrData        = 2
    OpNumeric         = 3
    OpString          = 4
    OpAccIndic        = 5
    OpAcc             = 6
    
    OpFormatDecimal   = 1
    OpFormatHexa      = 2
    OpFormatString    = 3
    OpFormatRegister  = 4
    
    OpType      = 0
    OpPattern   = 1
    OpFormat    = 2
    OpEquiv     = 3
 
    InstCode = 0
    InstFct  = 1
    InstOp   = 2
    
    def __init__(self, name, instructionSize, operandSize):
        self.name           = name
        self.loadAddress    = 0
        self.instructionSize = instructionSize
        self.operandSize= operandSize
        self.regList        = []
        self.indicatorList  = []
        self.endProgram = False
        Register.reset()
        Indicator.reset()
        
    @classmethod
    def whenUpdated(cls, fct):
        cls.displayChange = fct
        
    def addRegister(self, regName, regFormat, regType, regSize):
        reg = Register(self, regName, regFormat, regType, regSize)
        self.regList.append(reg)
        return reg
    
    def addIndicator(self, indicatorName, indicatorType):
        indicator = Indicator(self, indicatorName, indicatorType)
        self.indicatorList.append(indicator)
        return indicator
        
    def getRegisterList(self):
        return self.regList
        
    def getIndicatorList(self):
        return self.indicatorList
    
    def setZero(self, value):
        for ind in self.indicatorList:
            if ind.indicatorType == Indicator.Zero:
                if (value == 0):
                    ind.set(1)
                else:
                    ind.set(0)
                    
    def setNeg(self, value):
        for ind in self.indicatorList:
            if ind.indicatorType == Indicator.Negative:
                if (value < 0):
                    ind.set(1)
                else:
                    ind.set(0)
                    
    def setOverflow(self, value):
        for ind in self.indicatorList:
            if ind.indicatorType == Indicator.Overflow:
                ind.set(value)
                
    def clearIndicator(self):
        for ind in self.indicatorList:
            ind.set(0)
            
    def defineOperand(self, operandSet):
        self.operandSet = operandSet
        
    def defineInstruction(self, instructionSet):
        self.instructionSet = instructionSet
        
    def executeNext(self, board):
        memory = board.memory
        Processor.displayChange(self.PC.get())
        address = self.PC.get()
        
        opcode = memory.get(address, self.instructionSize)
        self.PC.set(self.PC.get() + self.instructionSize)

        found = False
        for mnemonic in self.instructionSet.keys():
            if self.instructionSet[mnemonic][self.InstCode] == opcode:
                found = True
                break
        if (found):
            if (self.instructionSet[mnemonic][self.InstOp][0] == Processor.OpVoid):
                operand = None
            else:
                operand = memory.get(self.PC.get(), self.operandSize)
                self.PC.set(self.PC.get() + self.operandSize)
            self.instructionSet[mnemonic][self.InstFct](self, board, operand)
        else:
            raise ExecError(Error.error, "Address: " + str(hex(address)) + ": invalid opcode " + str(hex(opcode)))
            return True
        
        return self.endProgram
                
class Assembly:
        
    def __init__(self):
        self.code = 0;
        self.loadAddress = 0
        self.instructionList = []
            
    def storeInstruction(self, address, mnemonic, operand):
        self.instructionList.append([address, mnemonic, operand])
        
    def getInstructionList(self):
        return self.instructionList

    def setCodeBase(self, value):
        self.code = value
        self.loadAddress = value
        
    def getCodeBase(self):
        return self.code
    
    def getLoadAddress(self):
        return self.loadAddress
    
    def loadAssembly(self,board, mnemonic, operand):
        
        memory = board.memory
        
        errorMsg = ""
        if (mnemonic in board.chip.instructionSet):

            memory.set(self.loadAddress, board.chip.instructionSize, board.chip.instructionSet[mnemonic][board.chip.InstCode])
            self.loadAddress += board.chip.instructionSize

            if operand:
                # determine the operand type
                # note that we look for all possible type of operands (several can match)
                foundOp = []
                for op in board.chip.operandSet:
                    p = re.compile("^" + op[Processor.OpPattern] + "$") 
                    if p.match(operand):
                        foundOp.append(op)
                            
                if (len(foundOp) == 0):
                    errorMsg="unknown operand type " + operand
                else:
                    # check if one of the found pattern is authorized
                    found = False
                    for op in foundOp:
                        opType = op[Processor.OpType]
                        if opType in board.chip.instructionSet[mnemonic][Processor.InstOp]:
                            found = True
                            opEquiv = op[Processor.OpEquiv]
                            opFormat = op[Processor.OpFormat]
                            opPattern = "^" + op[Processor.OpPattern] + "$"
                            break
                    
                    if (not found):
                        errorMsg="operand " + operand + " not valid for " + mnemonic
                    else:
                        error = False
                        if opEquiv == -1:
                            p = re.match(opPattern, operand)
                            if p:
                                realOperand = p.group(1)
                            else:
                                error = True
                                errorMsg="internal error: checking pattern matching"

                            if opFormat == Processor.OpFormatDecimal:
                                print "realOperand : " + str(realOperand)
                                value = int(realOperand)
                                if (value < 0):
                                    print " -----------------------> NEGATIF"
                                    print "   -> realOperand: *" + str(realOperand) + "*"
                                    print "    -> value: " + str(value)
                                
                            elif opFormat == Processor.OpFormatHexa:
                                value = int (realOperand, 16)
                            else:
                                error = True
                                errorMsg="internal error: format " + opFormat + " not implemented"
                            if not error:
                                memory.set(self.loadAddress, board.chip.operandSize, value)
                                self.loadAddress += board.chip.operandSize
 
                        else:
                            memory.set(self.loadAddress, board.chip.operandSize, opEquiv)
                            self.loadAddress += board.chip.operandSize
            else:
                if (Processor.OpVoid not in board.chip.instructionSet[mnemonic][Processor.InstOp]):
                    errorMsg="operand missing"
                    memory.set(self.loadAddress, board.chip.operandSize, 0)
                    self.loadAddress += board.chip.operandSize
        else:
            errorMsg="Unknown instruction" + mnemonic
            
        return errorMsg
